#
# Common tests for netplan OpenVSwitch support
#
# Copyright (C) 2020 Canonical, Ltd.
# Author: Łukasz 'sil2100' Zemczak <lukasz.zemczak@ubuntu.com>
#         Lukas 'slyon' Märdian <lukas.maerdian@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

from netplan_cli.cli.ovs import OVS_VSCTL_PATH
from .base import TestBase, ND_EMPTY, ND_WITHIP, ND_DHCP4, ND_DHCP6, \
                            OVS_PHYSICAL, OVS_VIRTUAL, \
                            OVS_BR_EMPTY, OVS_BR_DEFAULT, \
                            OVS_CLEANUP


@unittest.skipIf(not os.path.exists(OVS_VSCTL_PATH),
                 'OpenVSwitch not installed')
class TestOpenVSwitch(TestBase):
    '''OVS output'''

    def test_interface_external_ids_other_config(self):
        self.generate('''network:
  version: 2
  bridges: # bridges first, to trigger multi-pass processing
    ovs0:
      interfaces: [eth0, eth1]
      openvswitch: {}
  ethernets:
    eth0:
      openvswitch:
        external-ids:
          iface-id: myhostname
        other-config:
          disable-in-band: true
      dhcp6: true
    eth1:
      dhcp4: true
      openvswitch:
        other-config:
          disable-in-band: false\n''')
        self.assert_ovs({'ovs0.service': OVS_VIRTUAL % {'iface': 'ovs0', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br ovs0
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port ovs0 eth1
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port ovs0 eth0
''' + OVS_BR_DEFAULT % {'iface': 'ovs0'}},
                         'eth0.service': OVS_PHYSICAL % {'iface': 'eth0', 'extra': '''\
Requires=netplan-ovs-ovs0.service
After=netplan-ovs-ovs0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth0 external-ids:iface-id="myhostname"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth0 external-ids:netplan/external-ids/iface-id="myhostname"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth0 other-config:disable-in-band="true"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth0 external-ids:netplan/other-config/disable-in-band="true"
'''},
                         'eth1.service': OVS_PHYSICAL % {'iface': 'eth1', 'extra': '''\
Requires=netplan-ovs-ovs0.service
After=netplan-ovs-ovs0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth1 other-config:disable-in-band="false"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth1 external-ids:netplan/other-config/disable-in-band="false"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'ovs0.network': ND_EMPTY % ('ovs0', 'ipv6'),
                              'eth0.network': (ND_DHCP6 % 'eth0')
                              .replace('LinkLocalAddressing=ipv6', 'LinkLocalAddressing=no\nBridge=ovs0'),
                              'eth1.network': (ND_DHCP4 % 'eth1')
                              .replace('LinkLocalAddressing=ipv6', 'LinkLocalAddressing=no\nBridge=ovs0')})

    def test_interface_invalid_external_ids_other_config(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth0:
      openvswitch:
        external-ids:
          iface-id: myhostname
        other-config:
          disable-in-band: true''', expect_fail=True)
        self.assertIn('eth0: Interface needs to be assigned to an OVS bridge/bond to carry external-ids/other-config', err)

    def test_global_external_ids_other_config(self):
        self.generate('''network:
  version: 2
  openvswitch:
    external-ids:
      iface-id: myhostname
    other-config:
      disable-in-band: true
  ethernets:
    eth0:
      dhcp4: yes
''')
        self.assert_ovs({'global.service': OVS_VIRTUAL % {'iface': 'global', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' set open_vswitch . external-ids:iface-id="myhostname"
ExecStart=''' + OVS_VSCTL_PATH + ''' set open_vswitch . external-ids:netplan/external-ids/iface-id="myhostname"
ExecStart=''' + OVS_VSCTL_PATH + ''' set open_vswitch . other-config:disable-in-band="true"
ExecStart=''' + OVS_VSCTL_PATH + ''' set open_vswitch . external-ids:netplan/other-config/disable-in-band="true"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'eth0.network': ND_DHCP4 % 'eth0'})

    def test_global_set_protocols(self):
        out = self.generate('''network:
  version: 2
  openvswitch:
    protocols: [OpenFlow10, OpenFlow11, OpenFlow12, OpenFlow16]
  bridges:
    ovs0:
      openvswitch: {}''', skip_generated_yaml_validation=True)  # OpenFlow16 won't be re-generated
        self.assertIn('Open vSwitch: Ignoring deprecated protocol: OpenFlow16', out)
        self.assert_ovs({'ovs0.service': OVS_VIRTUAL % {'iface': 'ovs0', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br ovs0
''' + OVS_BR_DEFAULT % {'iface': 'ovs0'} + '''\
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge ovs0 protocols=OpenFlow10,OpenFlow11,OpenFlow12
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge ovs0 external-ids:netplan/protocols="OpenFlow10,OpenFlow11,OpenFlow12"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'ovs0.network': ND_EMPTY % ('ovs0', 'ipv6')})

    def test_duplicate_map_entry(self):
        err = self.generate('''network:
  version: 2
  openvswitch:
    external-ids:
      iface-id: myhostname
      iface-id: foobar
  ethernets:
    eth0:
      dhcp4: yes
''', expect_fail=True)
        self.assertIn("duplicate map entry 'iface-id'", err)

    def test_no_ovs_config(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth0:
      dhcp4: yes
''')
        self.assert_ovs({'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        self.assert_networkd({'eth0.network': ND_DHCP4 % 'eth0'})

    def test_bond_setup(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth1: {}
    eth2: {}
  bonds:
    bond0:
      interfaces: [eth1, eth2]
      openvswitch:
        external-ids:
          iface-id: myhostname
  bridges:
    br0:
      addresses: [192.170.1.1/24]
      interfaces: [bond0]
      openvswitch: {}
''')
        self.assert_ovs({'bond0.service': OVS_VIRTUAL % {'iface': 'bond0', 'extra':
                        '''Requires=netplan-ovs-br0.service
After=netplan-ovs-br0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-bond br0 bond0 eth1 eth2
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan="true"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 lacp=off
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan/lacp="off"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:iface-id="myhostname"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan/external-ids/iface-id="myhostname"
'''},
                         'br0.service': OVS_BR_EMPTY % {'iface': 'br0'},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'eth1.network': '[Match]\nName=eth1\n\n[Network]\nLinkLocalAddressing=no\nBond=bond0\n',
                              'eth2.network': '[Match]\nName=eth2\n\n[Network]\nLinkLocalAddressing=no\nBond=bond0\n',
                              'br0.network': ND_WITHIP % ('br0', '192.170.1.1/24'),
                              'bond0.network': ND_EMPTY % ('bond0', 'no')})

    def test_bond_no_bridge(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth1: {}
    eth2: {}
  bonds:
    bond0:
      interfaces: [eth1, eth2]
      openvswitch: {}
''', expect_fail=True)
        self.assertIn("Bond bond0 needs to be a member of an OpenVSwitch bridge", err)

    def test_bond_not_enough_interfaces(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth1: {}
  bonds:
    bond0:
      interfaces: [eth1]
      openvswitch: {}
  bridges:
    br0:
      addresses: [192.170.1.1/24]
      interfaces: [bond0]
      openvswitch: {}
''', expect_fail=True)
        self.assertIn("Bond bond0 needs to have at least 2 member interfaces", err)

    def test_bond_lacp(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth1: {}
    eth2: {}
  bonds:
    bond0:
      interfaces: [eth1, eth2]
      openvswitch:
        lacp: active
  bridges:
    br0:
      addresses: [192.170.1.1/24]
      interfaces: [bond0]
      openvswitch: {}
''')
        self.assert_ovs({'bond0.service': OVS_VIRTUAL % {'iface': 'bond0', 'extra':
                        '''Requires=netplan-ovs-br0.service
After=netplan-ovs-br0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-bond br0 bond0 eth1 eth2
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan="true"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 lacp=active
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan/lacp="active"
'''},
                         'br0.service': OVS_BR_EMPTY % {'iface': 'br0'},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'eth1.network': '[Match]\nName=eth1\n\n[Network]\nLinkLocalAddressing=no\nBond=bond0\n',
                              'eth2.network': '[Match]\nName=eth2\n\n[Network]\nLinkLocalAddressing=no\nBond=bond0\n',
                              'br0.network': ND_WITHIP % ('br0', '192.170.1.1/24'),
                              'bond0.network': ND_EMPTY % ('bond0', 'no')})

    def test_bond_lacp_invalid(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth1: {}
    eth2: {}
  bonds:
    bond0:
      interfaces: [eth1, eth2]
      openvswitch:
        lacp: invalid
  bridges:
    br0:
      addresses: [192.170.1.1/24]
      interfaces: [bond0]
      openvswitch: {}
''', expect_fail=True)
        self.assertIn("Value of 'lacp' needs to be 'active', 'passive' or 'off", err)

    def test_bond_lacp_wrong_type(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth1:
      openvswitch:
        lacp: passive
''', expect_fail=True)
        self.assertIn("Key 'lacp' is only valid for interface type 'Open vSwitch bond'", err)

    def test_bond_mode_implicit_params(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth1: {}
    eth2: {}
  bonds:
    bond0:
      interfaces: [eth1, eth2]
      parameters:
        mode: balance-tcp # Sets OVS backend implicitly
  bridges:
    br0:
      addresses: [192.170.1.1/24]
      interfaces: [bond0]
      openvswitch: {}
''')
        self.assert_ovs({'bond0.service': OVS_VIRTUAL % {'iface': 'bond0', 'extra':
                        '''Requires=netplan-ovs-br0.service
After=netplan-ovs-br0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-bond br0 bond0 eth1 eth2
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan="true"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 lacp=off
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan/lacp="off"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 bond_mode=balance-tcp
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan/bond_mode="balance-tcp"
'''},
                         'br0.service': OVS_BR_EMPTY % {'iface': 'br0'},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'eth1.network': '[Match]\nName=eth1\n\n[Network]\nLinkLocalAddressing=no\nBond=bond0\n',
                              'eth2.network': '[Match]\nName=eth2\n\n[Network]\nLinkLocalAddressing=no\nBond=bond0\n',
                              'br0.network': ND_WITHIP % ('br0', '192.170.1.1/24'),
                              'bond0.network': ND_EMPTY % ('bond0', 'no')})

    def test_bond_mode_explicit_params(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth1: {}
    eth2: {}
  bonds:
    bond0:
      interfaces: [eth1, eth2]
      parameters:
        mode: active-backup
      openvswitch: {}
  bridges:
    br0:
      addresses: [192.170.1.1/24]
      interfaces: [bond0]
      openvswitch: {}
''')
        self.assert_ovs({'bond0.service': OVS_VIRTUAL % {'iface': 'bond0', 'extra':
                        '''Requires=netplan-ovs-br0.service
After=netplan-ovs-br0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-bond br0 bond0 eth1 eth2
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan="true"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 lacp=off
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan/lacp="off"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 bond_mode=active-backup
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan/bond_mode="active-backup"
'''},
                         'br0.service': OVS_BR_EMPTY % {'iface': 'br0'},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'eth1.network': '[Match]\nName=eth1\n\n[Network]\nLinkLocalAddressing=no\nBond=bond0\n',
                              'eth2.network': '[Match]\nName=eth2\n\n[Network]\nLinkLocalAddressing=no\nBond=bond0\n',
                              'br0.network': ND_WITHIP % ('br0', '192.170.1.1/24'),
                              'bond0.network': ND_EMPTY % ('bond0', 'no')})

    def test_bond_mode_ovs_invalid(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth1: {}
    eth2: {}
  bonds:
    bond0:
      interfaces: [eth1, eth2]
      parameters:
        mode: balance-rr
      openvswitch: {}
  bridges:
    br0:
      addresses: [192.170.1.1/24]
      interfaces: [bond0]
      openvswitch: {}
''', expect_fail=True)
        self.assertIn("bond0: bond mode 'balance-rr' not supported by Open vSwitch", err)

    def test_bridge_setup(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth1: {}
    eth2: {}
  bridges:
    br0:
      addresses: [192.170.1.1/24]
      interfaces: [eth1, eth2]
      openvswitch: {}
''')
        self.assert_ovs({'br0.service': OVS_VIRTUAL % {'iface': 'br0', 'extra':
                        '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br0
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port br0 eth1
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port br0 eth2
''' + OVS_BR_DEFAULT % {'iface': 'br0'}},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'eth1.network': '[Match]\nName=eth1\n\n[Network]\nLinkLocalAddressing=no\nBridge=br0\n',
                              'eth2.network': '[Match]\nName=eth2\n\n[Network]\nLinkLocalAddressing=no\nBridge=br0\n',
                              'br0.network': ND_WITHIP % ('br0', '192.170.1.1/24')})

    def test_bridge_external_ids_other_config(self):
        self.generate('''network:
  version: 2
  bridges:
    br0:
      openvswitch:
        external-ids:
          iface-id: myhostname
        other-config:
          disable-in-band: true
''')
        self.assert_ovs({'br0.service': OVS_VIRTUAL % {'iface': 'br0', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br0
''' + OVS_BR_DEFAULT % {'iface': 'br0'} + '''\
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 external-ids:iface-id="myhostname"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 external-ids:netplan/external-ids/iface-id="myhostname"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 other-config:disable-in-band="true"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 external-ids:netplan/other-config/disable-in-band="true"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the bridge has been only configured for OVS
        self.assert_networkd({'br0.network': ND_EMPTY % ('br0', 'ipv6')})

    def test_bridge_non_default_parameters(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth1: {}
    eth2: {}
  bridges:
    br0:
      addresses: [192.170.1.1/24]
      interfaces: [eth1, eth2]
      openvswitch:
        fail-mode: secure
        mcast-snooping: true
        rstp: true
''')
        self.assert_ovs({'br0.service': OVS_VIRTUAL % {'iface': 'br0', 'extra':
                        '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br0
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port br0 eth1
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port br0 eth2
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 external-ids:netplan="true"
ExecStart=''' + OVS_VSCTL_PATH + ''' set-fail-mode br0 secure
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 external-ids:netplan/global/set-fail-mode="secure"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 mcast_snooping_enable=true
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 external-ids:netplan/mcast_snooping_enable="true"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 rstp_enable=true
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 external-ids:netplan/rstp_enable="true"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'eth1.network': '[Match]\nName=eth1\n\n[Network]\nLinkLocalAddressing=no\nBridge=br0\n',
                              'eth2.network': '[Match]\nName=eth2\n\n[Network]\nLinkLocalAddressing=no\nBridge=br0\n',
                              'br0.network': ND_WITHIP % ('br0', '192.170.1.1/24')})

    def test_bridge_fail_mode_invalid(self):
        err = self.generate('''network:
  version: 2
  bridges:
    br0:
      openvswitch:
        fail-mode: glorious
''', expect_fail=True)
        self.assertIn("Value of 'fail-mode' needs to be 'standalone' or 'secure'", err)

    def test_fail_mode_non_bridge(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth0:
      openvswitch:
        fail-mode: glorious
''', expect_fail=True)
        self.assertIn("Key 'fail-mode' is only valid for interface type 'Open vSwitch bridge'", err)

    def test_rstp_non_bridge(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth0:
      openvswitch:
        rstp: true
''', expect_fail=True)
        self.assertIn("Key is only valid for interface type 'Open vSwitch bridge'", err)

    def test_bridge_set_protocols(self):
        self.generate('''network:
  version: 2
  bridges:
    br0:
      openvswitch:
        protocols: [OpenFlow10, OpenFlow11, OpenFlow15]
''')
        self.assert_ovs({'br0.service': OVS_VIRTUAL % {'iface': 'br0', 'extra':
                        '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br0
''' + OVS_BR_DEFAULT % {'iface': 'br0'} + '''\
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 protocols=OpenFlow10,OpenFlow11,OpenFlow15
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 external-ids:netplan/protocols="OpenFlow10,OpenFlow11,OpenFlow15"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'br0.network': ND_EMPTY % ('br0', 'ipv6')})

    def test_bridge_set_protocols_invalid(self):
        err = self.generate('''network:
  version: 2
  bridges:
    br0:
      openvswitch:
        protocols: [OpenFlow10, OpenFooBar13, OpenFlow15]
''', expect_fail=True)
        self.assertIn("Unsupported OVS 'protocol' value: OpenFooBar13", err)

    def test_set_protocols_invalid_interface(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth0:
      openvswitch:
        protocols: [OpenFlow10, OpenFlow15]
''', expect_fail=True)
        self.assertIn("Key 'protocols' is only valid for interface type 'Open vSwitch bridge'", err)

    def test_bridge_controller(self):
        self.generate('''network:
  version: 2
  bridges:
    br0:
      openvswitch:
        controller:
          addresses: ["ptcp:", "ptcp:1337", "ptcp:1337:[fe80::1234%eth0]", "pssl:1337:[fe80::1]", "ssl:10.10.10.1",\
                      tcp:127.0.0.1:1337, "tcp:[fe80::1234%eth0]", "tcp:[fe80::1]:1337", unix:/some/path, punix:other/path]
          connection-mode: out-of-band
  openvswitch:
    ssl:
      ca-cert: /another/path
      certificate: /some/path
      private-key: /key/path
''')
        self.assert_ovs({'br0.service': OVS_VIRTUAL % {'iface': 'br0', 'extra':
                        '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br0
''' + OVS_BR_DEFAULT % {'iface': 'br0'} + '''\
ExecStart=''' + OVS_VSCTL_PATH + ''' set-controller br0 ptcp: ptcp:1337 ptcp:1337:[fe80::1234%eth0] pssl:1337:[fe80::1] \
ssl:10.10.10.1 tcp:127.0.0.1:1337 tcp:[fe80::1234%eth0] tcp:[fe80::1]:1337 unix:/some/path punix:other/path
ExecStart=''' + OVS_VSCTL_PATH + ''' set Bridge br0 external-ids:netplan/global/set-controller="ptcp:,ptcp:1337,\
ptcp:1337:[fe80::1234%eth0],pssl:1337:[fe80::1],ssl:10.10.10.1,tcp:127.0.0.1:1337,tcp:[fe80::1234%eth0],tcp:[fe80::1]:1337,\
unix:/some/path,punix:other/path"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Controller br0 connection-mode=out-of-band
ExecStart=''' + OVS_VSCTL_PATH + ''' set Controller br0 external-ids:netplan/connection-mode="out-of-band"
'''},
                         'global.service': OVS_VIRTUAL % {'iface': 'global', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' set-ssl /key/path /some/path /another/path
ExecStart=''' + OVS_VSCTL_PATH + ''' set open_vswitch . external-ids:netplan/global/set-ssl="/key/path,/some/path,/another/path"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'br0.network': ND_EMPTY % ('br0', 'ipv6')})

    def test_bridge_controller_invalid_target(self):
        err = self.generate('''network:
  version: 2
  bridges:
    br0:
      openvswitch:
        controller:
          addresses: [ptcp]
''', expect_fail=True)
        self.assertIn("Unsupported OVS controller target: ptcp", err)
        self.assert_ovs({})
        self.assert_networkd({})

    def test_bridge_controller_invalid_target_ip(self):
        err = self.generate('''network:
  version: 2
  bridges:
    br0:
      openvswitch:
        controller:
          addresses: ["tcp:[fe80:1234%eth0]"]
''', expect_fail=True)
        self.assertIn("Unsupported OVS controller target: tcp:[fe80:1234%eth0]", err)
        self.assert_ovs({})
        self.assert_networkd({})

    def test_bridge_controller_invalid_target_port(self):
        err = self.generate('''network:
  version: 2
  bridges:
    br0:
      openvswitch:
        controller:
          addresses: [ptcp:65536]
''', expect_fail=True)
        self.assertIn("Unsupported OVS controller target: ptcp:65536", err)
        self.assert_ovs({})
        self.assert_networkd({})

    def test_bridge_controller_invalid_connection_mode(self):
        err = self.generate('''network:
  version: 2
  bridges:
    br0:
      openvswitch:
        controller:
          connection-mode: INVALID
''', expect_fail=True)
        self.assertIn("Value of 'connection-mode' needs to be 'in-band' or 'out-of-band'", err)
        self.assert_ovs({})
        self.assert_networkd({})

    def test_bridge_controller_connection_mode_invalid_interface_type(self):
        err = self.generate('''network:
  version: 2
  bonds:
    mybond:
      openvswitch:
        controller:
          connection-mode: in-band
''', expect_fail=True)
        self.assertIn("Key 'controller.connection-mode' is only valid for interface type 'Open vSwitch bridge'", err)
        self.assert_ovs({})
        self.assert_networkd({})

    def test_bridge_controller_addresses_invalid_interface_type(self):
        err = self.generate('''network:
  version: 2
  bonds:
    mybond:
      openvswitch:
        controller:
          addresses: [unix:/some/socket]
''', expect_fail=True)
        self.assertIn("Key 'controller.addresses' is only valid for interface type 'Open vSwitch bridge'", err)
        self.assert_ovs({})
        self.assert_networkd({})

    def test_global_ssl(self):
        self.generate('''network:
  version: 2
  openvswitch:
    ssl:
      ca-cert: /another/path
      certificate: /some/path
      private-key: /key/path
''')
        self.assert_ovs({'global.service': OVS_VIRTUAL % {'iface': 'global', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' set-ssl /key/path /some/path /another/path
ExecStart=''' + OVS_VSCTL_PATH + ''' set open_vswitch . external-ids:netplan/global/set-ssl="/key/path,/some/path,/another/path"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({})

    def test_missing_ssl(self):
        err = self.generate('''network:
  version: 2
  bridges:
    br0:
      openvswitch:
        controller:
          addresses: [ssl:10.10.10.1]
  openvswitch:
    ssl: {}
''', expect_fail=True)
        self.assertIn("ERROR: Open vSwitch bridge controller target 'ssl:10.10.10.1' needs SSL configuration, but global \
'openvswitch.ssl' settings are not set", err)
        self.assert_ovs({'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        self.assert_networkd({})

    def test_global_ports(self):
        err = self.generate('''network:
  version: 2
  openvswitch:
    ports:
      - [patch0-1, patch1-0]
''', expect_fail=True)
        self.assertIn('patch0-1: OpenVSwitch patch port needs to be assigned to a bridge/bond', err)
        self.assert_ovs({'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        self.assert_networkd({})

    def test_few_ports(self):
        err = self.generate('''network:
  version: 2
  openvswitch:
    ports:
      - [patch0-1]
''', expect_fail=True)
        self.assertIn("An Open vSwitch peer port sequence must have exactly two entries", err)
        self.assertIn("- [patch0-1]", err)
        self.assert_ovs({})
        self.assert_networkd({})

    def test_many_ports(self):
        err = self.generate('''network:
  version: 2
  openvswitch:
    ports:
      - [patch0-1, "patchx", patchy]
''', expect_fail=True)
        self.assertIn("An Open vSwitch peer port sequence must have exactly two entries", err)
        self.assertIn("- [patch0-1, \"patchx\", patchy]", err)
        self.assert_ovs({})
        self.assert_networkd({})

    def test_ovs_invalid_port(self):
        err = self.generate('''network:
  version: 2
  openvswitch:
    ports:
      - [patchx, patchy]
      - [patchx, patchz]
''', expect_fail=True)
        self.assertIn("Open vSwitch port 'patchx' is already assigned to peer 'patchy'", err)
        self.assert_ovs({})
        self.assert_networkd({})

    def test_ovs_invalid_peer(self):
        err = self.generate('''network:
  version: 2
  openvswitch:
    ports:
      - [patchx, patchy]
      - [patchz, patchx]
''', expect_fail=True)
        self.assertIn("Open vSwitch port 'patchx' is already assigned to peer 'patchy'", err)
        self.assert_ovs({})
        self.assert_networkd({})

    def test_bridge_auto_ovs_backend(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth1: {}
    eth2: {}
  bonds:
    bond0:
      interfaces: [eth1, eth2]
      openvswitch: {}
  bridges:
    br0:
      addresses: [192.170.1.1/24]
      interfaces: [bond0]
''')
        self.assert_ovs({'br0.service': OVS_BR_EMPTY % {'iface': 'br0'},
                         'bond0.service': OVS_VIRTUAL % {'iface': 'bond0', 'extra':
                                                         '''Requires=netplan-ovs-br0.service
After=netplan-ovs-br0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-bond br0 bond0 eth1 eth2
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan="true"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 lacp=off
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan/lacp="off"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        self.assert_networkd({'br0.network': ND_WITHIP % ('br0', '192.170.1.1/24'),
                              'bond0.network': ND_EMPTY % ('bond0', 'no'),
                              'eth1.network':
                              '''[Match]
Name=eth1

[Network]
LinkLocalAddressing=no
Bond=bond0
''',
                              'eth2.network':
                              '''[Match]
Name=eth2

[Network]
LinkLocalAddressing=no
Bond=bond0
'''})

    def test_bond_auto_ovs_backend(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth0: {}
  bonds:
    bond0:
      interfaces: [eth0, patchy]
  bridges:
    br0:
      addresses: [192.170.1.1/24]
      interfaces: [bond0]
    br1:
      addresses: [2001:FFfe::1/64]
      interfaces: [patchx]
  openvswitch:
    ports:
      - [patchx, patchy]
''')
        self.assert_ovs({'br0.service': OVS_BR_EMPTY % {'iface': 'br0'},
                         'br1.service': OVS_VIRTUAL % {'iface': 'br1', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br1
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port br1 patchx -- set Interface patchx type=patch options:peer=patchy
''' + OVS_BR_DEFAULT % {'iface': 'br1'}},
                         'bond0.service': OVS_VIRTUAL % {'iface': 'bond0', 'extra':
                                                         '''Requires=netplan-ovs-br0.service
After=netplan-ovs-br0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-bond br0 bond0 patchy eth0 -- set Interface patchy type=patch \
options:peer=patchx
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan="true"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 lacp=off
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port bond0 external-ids:netplan/lacp="off"
'''},
                         'patchx.service': OVS_VIRTUAL % {'iface': 'patchx', 'extra':
                                                          '''Requires=netplan-ovs-br1.service
After=netplan-ovs-br1.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port patchx external-ids:netplan="true"
'''},
                         'patchy.service': OVS_VIRTUAL % {'iface': 'patchy', 'extra':
                                                          '''Requires=netplan-ovs-bond0.service
After=netplan-ovs-bond0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface patchy external-ids:netplan="true"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        self.assert_networkd({'br0.network': ND_WITHIP % ('br0', '192.170.1.1/24'),
                              'br1.network': ND_WITHIP % ('br1', '2001:FFfe::1/64'),
                              'bond0.network': ND_EMPTY % ('bond0', 'no'),
                              'patchx.network': ND_EMPTY % ('patchx', 'no'),
                              'patchy.network': ND_EMPTY % ('patchy', 'no'),
                              'eth0.network': '[Match]\nName=eth0\n\n[Network]\nLinkLocalAddressing=no\nBond=bond0\n'})

    def test_patch_ports(self):
        self.generate('''network:
  version: 2
  openvswitch:
    ports:
      - [patch0-1, patch1-0]
  bridges:
    br0:
      addresses: [192.168.1.1/24]
      interfaces: [patch0-1]
    br1:
      addresses: [192.168.1.2/24]
      interfaces: [patch1-0]
''')
        self.assert_ovs({'br0.service': OVS_VIRTUAL % {'iface': 'br0', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br0
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port br0 patch0-1 -- set Interface patch0-1 type=patch options:peer=patch1-0
''' + OVS_BR_DEFAULT % {'iface': 'br0'}},
                         'br1.service': OVS_VIRTUAL % {'iface': 'br1', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br1
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port br1 patch1-0 -- set Interface patch1-0 type=patch options:peer=patch0-1
''' + OVS_BR_DEFAULT % {'iface': 'br1'}},
                         'patch0-1.service': OVS_VIRTUAL % {'iface': 'patch0-1', 'extra':
                                                            '''Requires=netplan-ovs-br0.service
After=netplan-ovs-br0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port patch0-1 external-ids:netplan="true"
'''},
                         'patch1-0.service': OVS_VIRTUAL % {'iface': 'patch1-0', 'extra':
                                                            '''Requires=netplan-ovs-br1.service
After=netplan-ovs-br1.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' set Port patch1-0 external-ids:netplan="true"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        self.assert_networkd({'br0.network': ND_WITHIP % ('br0', '192.168.1.1/24'),
                              'br1.network': ND_WITHIP % ('br1', '192.168.1.2/24'),
                              'patch0-1.network': ND_EMPTY % ('patch0-1', 'no'),
                              'patch1-0.network': ND_EMPTY % ('patch1-0', 'no')})

    def test_fake_vlan_bridge_setup(self):
        self.generate('''network:
  version: 2
  bridges:
    br0:
      addresses: [192.168.1.1/24]
      openvswitch: {}
  vlans:
    br0.100:
      id: 100
      link: br0
      openvswitch: {}
''')
        self.assert_ovs({'br0.service': OVS_VIRTUAL % {'iface': 'br0', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br0
''' + OVS_BR_DEFAULT % {'iface': 'br0'}},
                         'br0.100.service': OVS_VIRTUAL % {'iface': 'br0.100', 'extra':
                                                           '''Requires=netplan-ovs-br0.service
After=netplan-ovs-br0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br0.100 br0 100
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface br0.100 external-ids:netplan="true"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'br0.network': ND_WITHIP % ('br0', '192.168.1.1/24'),
                              'br0.100.network': ND_EMPTY % ('br0.100', 'ipv6')})

    def test_implicit_fake_vlan_bridge_setup(self):
        # Test if, when a VLAN is added to an OVS bridge, netplan will
        # implicitly assume the vlan should be done via OVS as well
        self.generate('''network:
  version: 2
  bridges:
    br0:
      addresses: [192.168.1.1/24]
      openvswitch: {}
  vlans:
    br0.100:
      id: 100
      link: br0
''')
        self.assert_ovs({'br0.service': OVS_BR_EMPTY % {'iface': 'br0'},
                         'br0.100.service': OVS_VIRTUAL % {'iface': 'br0.100', 'extra':
                                                           '''Requires=netplan-ovs-br0.service
After=netplan-ovs-br0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br0.100 br0 100
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface br0.100 external-ids:netplan="true"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'br0.network': ND_WITHIP % ('br0', '192.168.1.1/24'),
                              'br0.100.network': ND_EMPTY % ('br0.100', 'ipv6')})

    def test_invalid_device_type(self):
        err = self.generate('''network:
    version: 2
    ethernets:
        eth0:
            openvswitch: {}
''', expect_fail=True)
        self.assertIn('eth0: This device type is not supported with the OpenVSwitch backend', err)
        self.assert_ovs({'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        self.assert_networkd({})

    def test_bridge_non_ovs_bond(self):
        self.generate('''network:
    version: 2
    ethernets:
        eth0: {}
        eth1: {}
    bonds:
        non-ovs-bond:
            interfaces: [eth0, eth1]
    bridges:
        ovs-br:
            interfaces: [non-ovs-bond]
            openvswitch: {}
''')
        self.assert_ovs({'ovs-br.service': OVS_VIRTUAL % {'iface': 'ovs-br', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br ovs-br
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port ovs-br non-ovs-bond
''' + OVS_BR_DEFAULT % {'iface': 'ovs-br'}},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'non-ovs-bond.network': ND_EMPTY % ('non-ovs-bond', 'no') + 'Bridge=ovs-br\n',
                              'eth1.network': (ND_EMPTY % ('eth1', 'no')).replace('ConfigureWithoutCarrier=yes',
                              'Bond=non-ovs-bond'),
                              'eth0.network': (ND_EMPTY % ('eth0', 'no')).replace('ConfigureWithoutCarrier=yes',
                              'Bond=non-ovs-bond'),
                              'ovs-br.network': ND_EMPTY % ('ovs-br', 'ipv6'),
                              'non-ovs-bond.netdev': '[NetDev]\nName=non-ovs-bond\nKind=bond\n'})

    def test_ovs_invalid_networkd_config(self):
        err = self.generate('''network:
  version: 2
  bridges:
    br0:
      openvswitch: {}
      dhcp4: true
      dhcp4-overrides:
        use-domains: true
      dhcp6: true
      dhcp6-overrides:
        use-domains: false
''', expect_fail=True)
        self.assert_ovs({'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        self.assertIn('br0: networkd requires that use-domains has the same value', err)

    def test_ovs_duplicates_when_parser_needs_second_pass(self):
        ''' Test case for LP: #2007682
            The generator shouldn't generate duplicates when the parser
            needs a second pass.
        '''
        self.generate('''network:
    version: 2
    bridges:
      br123:
        openvswitch:
          protocols:
            - OpenFlow10
            - OpenFlow11
            - OpenFlow12
          controller:
            addresses:
              - tcp:127.0.0.1:6653
        interfaces:
          - nic1
    ethernets:
      nic1: {}
''')
        self.assert_ovs({'br123.service': OVS_VIRTUAL % {'iface': 'br123', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br br123
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port br123 nic1
''' + OVS_BR_DEFAULT % {'iface': 'br123'} + ('\
ExecStart=' + OVS_VSCTL_PATH + ' set Bridge br123 protocols=OpenFlow10,OpenFlow11,OpenFlow12\n\
ExecStart=' + OVS_VSCTL_PATH + ' set Bridge br123 external-ids:netplan/protocols="OpenFlow10,OpenFlow11,OpenFlow12"\n\
ExecStart=' + OVS_VSCTL_PATH + ' set-controller br123 tcp:127.0.0.1:6653\n\
ExecStart=' + OVS_VSCTL_PATH + ' set Bridge br123 external-ids:netplan/global/set-controller="tcp:127.0.0.1:6653"\n\
')},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})

    def test_both_ports_with_same_name(self):
        err = self.generate('''network:
    version: 2
    openvswitch:
      ports:
        - [portname, portname]
''', expect_fail=True)
        self.assertIn('Open vSwitch patch ports must be of different name', err)

    def test_file_paths_escaped(self):
        self.generate('''network:
  version: 2
  bridges:
    "abc/../../123":
      openvswitch: {}
  vlans:
    "abc/../../123.100":
      id: 100
      link: "abc/../../123"
''')
        self.assert_ovs({'abc%2F..%2F..%2F123.service': OVS_BR_EMPTY % {'iface': 'abc/../../123'},
                         'abc%2F..%2F..%2F123.100.service': OVS_VIRTUAL % {'iface': 'abc/../../123.100', 'extra':
                                                                           '''Requires=netplan-ovs-abc%2F..%2F..%2F123.service
After=netplan-ovs-abc%2F..%2F..%2F123.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br abc/../../123.100 abc/../../123 100
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface abc/../../123.100 external-ids:netplan="true"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})

    def test_control_characters_and_semicolons_escaping(self):
        self.generate('''network:
  version: 2
  bridges: # bridges first, to trigger multi-pass processing
    ovs0:
      interfaces: [eth0, eth1]
      openvswitch: {}
  ethernets:
    eth0:
      openvswitch:
        external-ids:
          "a\\n1\\ra": " ; a ; 1 ;a; ;b\\t;\\t3 ;\\ta\\t; 1"
        other-config:
          "a\\n1\\ra": " ; a ; 1 ;a; ;b\\t;\\t3 ;\\ta\\t; 1"
      dhcp6: true
    eth1:
      dhcp4: true
      openvswitch:
        other-config:
          disable-in-band: false\n''', skip_generated_yaml_validation=True)
        self.assert_ovs({'ovs0.service': OVS_VIRTUAL % {'iface': 'ovs0', 'extra': '''
[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-br ovs0
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port ovs0 eth1
ExecStart=''' + OVS_VSCTL_PATH + ''' --may-exist add-port ovs0 eth0
''' + OVS_BR_DEFAULT % {'iface': 'ovs0'}},
                         'eth0.service': OVS_PHYSICAL % {'iface': 'eth0', 'extra': '''\
Requires=netplan-ovs-ovs0.service
After=netplan-ovs-ovs0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth0 external-ids:a\\n1\\ra=" \\; a \\; 1 ;a; ;b\\t;\\t3 ;\\ta\\t; 1"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth0 \
external-ids:netplan/external-ids/a\\n1\\ra=",;,a,;,1,;a;,;b\\t;\\t3,;\\ta\\t;,1"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth0 other-config:a\\n1\\ra=" \\; a \\; 1 ;a; ;b\\t;\\t3 ;\\ta\\t; 1"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth0 \
external-ids:netplan/other-config/a\\n1\\ra=",;,a,;,1,;a;,;b\\t;\\t3,;\\ta\\t;,1"
'''},
                         'eth1.service': OVS_PHYSICAL % {'iface': 'eth1', 'extra': '''\
Requires=netplan-ovs-ovs0.service
After=netplan-ovs-ovs0.service

[Service]
Type=oneshot
TimeoutStartSec=10s
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth1 other-config:disable-in-band="false"
ExecStart=''' + OVS_VSCTL_PATH + ''' set Interface eth1 external-ids:netplan/other-config/disable-in-band="false"
'''},
                         'cleanup.service': OVS_CLEANUP % {'iface': 'cleanup'}})
        # Confirm that the networkd config is still sane
        self.assert_networkd({'ovs0.network': ND_EMPTY % ('ovs0', 'ipv6'),
                              'eth0.network': (ND_DHCP6 % 'eth0')
                              .replace('LinkLocalAddressing=ipv6', 'LinkLocalAddressing=no\nBridge=ovs0'),
                              'eth1.network': (ND_DHCP4 % 'eth1')
                              .replace('LinkLocalAddressing=ipv6', 'LinkLocalAddressing=no\nBridge=ovs0')})



#!/usr/bin/python3
#
# Copyright (C) 2018 Canonical, Ltd.
# Author: Mathieu Trudel-Lapierre <mathieu.trudel-lapierre@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''netplan try command line'''

import logging
import netplan
import os
import time
import shutil
import signal
import sys
import tempfile

from ...configmanager import ConfigManager
from .. import utils
from .apply import NetplanApply
from ... import terminal

# Keep a timeout long enough to allow the network to converge, 60 seconds may
# be slightly short given some complex configs, i.e. if STP must reconverge.
DEFAULT_INPUT_TIMEOUT = 120


class NetplanTry(utils.NetplanCommand):

    def __init__(self):
        super().__init__(command_id='try',
                         description='Try to apply a new netplan config to running '
                                     'system, with automatic rollback',
                         leaf=True)
        self.configuration_changed = False
        self.new_interfaces = None
        self.config_file = None
        self._config_manager = None
        self.t_settings = None
        self.t = None
        self._rootdir = os.environ.get('DBUS_TEST_NETPLAN_ROOT', '/')
        self._netplan_try_stamp = os.path.join(self._rootdir, self.try_ready_stamp)

    @property
    def config_manager(self):  # pragma: nocover (called by later commands)
        if not self._config_manager:
            self._config_manager = ConfigManager(prefix=self._rootdir)
        return self._config_manager

    def clear_ready_stamp(self):
        if os.path.isfile(self._netplan_try_stamp):
            os.remove(self._netplan_try_stamp)
            return True
        return False

    def touch_ready_stamp(self):
        os.makedirs(self._rootdir + '/run/netplan', mode=0o700, exist_ok=True)
        open(self._netplan_try_stamp, 'w').close()

    def run(self):  # pragma: nocover (requires user input)
        self.parser.add_argument('--config-file',
                                 help='Apply the config file in argument in addition to current configuration.')
        self.parser.add_argument('--timeout',
                                 type=int, default=DEFAULT_INPUT_TIMEOUT,
                                 help="Maximum number of seconds to wait for the user's confirmation")
        self.parser.add_argument('--state',
                                 help='Directory containing previous YAML configuration')

        self.func = self.command_try

        self.parse_args()
        self.run_command()

    def command_try(self):  # pragma: nocover (requires user input)
        if not self.is_revertable():
            sys.exit(os.EX_CONFIG)

        try:
            fd = sys.stdin.fileno()
            self.t = terminal.Terminal(fd)
            self.t.save(self.t_settings)

            # we really don't want to be interrupted while doing backup/revert operations
            signal.signal(signal.SIGINT, self._signal_handler)
            signal.signal(signal.SIGUSR1, self._signal_handler)

            self.backup()
            self.setup()

            NetplanApply().command_apply(run_generate=True, sync=True, exit_on_error=False, state_dir=self.state)

            # Touch stamp file, it is the signal (for netplan-dbus) that we're
            # ready to accept any Accept/Reject input (like SIGUSR1 or SIGTERM)
            self.touch_ready_stamp()
            self.t.get_confirmation_input(timeout=self.timeout)
        except terminal.InputRejected:
            print("\nReverting.")
            self.revert()
        except terminal.InputAccepted:
            print("\nConfiguration accepted.")
        except Exception as e:
            print("\nAn error occurred: %s" % e)
            print("\nReverting.")
            self.revert()
        finally:
            if self.t:
                self.t.reset(self.t_settings)
            self.cleanup()
            self.clear_ready_stamp()

    def backup(self):  # pragma: nocover (requires user input)
        backup_config_dir = False
        if self.config_file:
            backup_config_dir = True
        self.config_manager.backup(backup_config_dir=backup_config_dir)

    def setup(self):  # pragma: nocover (requires user input)
        if self.config_file:
            dest_dir = os.path.join("/", "etc", "netplan")
            dest_name = os.path.basename(self.config_file).rstrip('.yaml')
            dest_suffix = time.time()
            dest_path = os.path.join(dest_dir, "{}.{}.yaml".format(dest_name, dest_suffix))
            self.config_manager.add({self.config_file: dest_path})
        self.configuration_changed = True

    def revert(self):  # pragma: nocover (requires user input)
        # backup the state we just tried to apply
        tempdir = tempfile.mkdtemp()
        confdir = os.path.join(tempdir, 'etc', 'netplan')
        os.makedirs(confdir)
        self.config_manager.copy_tree('/etc/netplan', confdir, dirs_exist_ok=True)
        # restore previous state
        self.config_manager.revert()
        NetplanApply().command_apply(run_generate=False, sync=True, exit_on_error=False, state_dir=tempdir)
        # clear the backup
        shutil.rmtree(tempdir)

    def cleanup(self):  # pragma: nocover (requires user input)
        self.config_manager.cleanup()

    def is_revertable(self):
        '''
        Check if the configuration is revertable, if it doesn't contain bits
        that we know are likely to render the system unstable if we apply it,
        or if we revert.

        Returns True if the parsed config is "revertable", meaning that we
        can actually rely on backends to re-apply /all/ of the relevant
        configuration to interfaces when their config changes.

        Returns False if the parsed config contains options that are known
        to not cleanly revert via the backend.
        '''

        extra_config = []
        if self.config_file:
            extra_config.append(self.config_file)
        np_state = None
        try:
            np_state = self.config_manager.parse(extra_config=extra_config)
        except utils.config_errors as e:
            logging.error(e)
            sys.exit(os.EX_CONFIG)

        revert_unsupported = []

        # Bridges and bonds are special. They typically include (or could include)
        # more than one device in them, and they can be set with special parameters
        # to tweak their behavior, which are really hard to "revert", especially
        # as systemd-networkd doesn't necessarily touch them when config changes.
        multi_iface: dict[str, netplan.NetDefinition] = {}
        multi_iface.update(np_state.bridges)
        multi_iface.update(np_state.bonds)
        for itf in multi_iface.values():
            if not itf._is_trivial_compound_itf:
                reason = "reverting custom parameters for bridges and bonds is not supported"
                revert_unsupported.append((itf.id, reason))

        if revert_unsupported:
            for ifname, reason in revert_unsupported:
                print("{}: {}".format(ifname, reason))
            print("\nPlease carefully review the configuration and use 'netplan apply' directly.")
            return False
        return True

    def _signal_handler(self, sig, frame):  # pragma: nocover (requires user input)
        if sig == signal.SIGUSR1:
            raise terminal.InputAccepted()
        else:
            if self.configuration_changed:
                raise terminal.InputRejected()



#
# Functional tests of netplan generate that verify that the generated
# configuration files look as expected. These are run during "make check" and
# don't touch the system configuration at all.
#
# Copyright (C) 2016-2021 Canonical, Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>
# Author: Lukas Märdian <slyon@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import random
import glob
import stat
import string
import tempfile
import subprocess
import unittest
import ctypes
import ctypes.util
import yaml
import difflib
import re
from io import StringIO

import netplan
from netplan_cli.cli.ovs import OVS_VSCTL_PATH

exe_generate = os.environ.get('NETPLAN_GENERATE_PATH',
                              os.path.join(os.path.dirname(os.path.dirname(
                                           os.path.dirname(os.path.abspath(__file__)))), 'generate'))
exe_configure = os.environ.get('NETPLAN_CONFIGURE_PATH',
                               os.path.join(os.path.dirname(os.path.dirname(
                                            os.path.dirname(os.path.abspath(__file__)))), 'configure'))

# make sure we point to libnetplan properly.
os.environ.update({'LD_LIBRARY_PATH': '.:{}'.format(os.environ.get('LD_LIBRARY_PATH'))})

# make sure we fail on criticals
os.environ['G_DEBUG'] = 'fatal-criticals'

lib = ctypes.CDLL('libnetplan.so.1')

# common patterns for expected output
ND_EMPTY = '[Match]\nName=%s\n\n[Network]\nLinkLocalAddressing=%s\nConfigureWithoutCarrier=yes\n'
ND_WITHIP = '[Match]\nName=%s\n\n[Network]\nLinkLocalAddressing=ipv6\nAddress=%s\nConfigureWithoutCarrier=yes\n'
ND_WIFI_DHCP4 = '[Match]\nName=%s\n\n[Network]\nDHCP=ipv4\nLinkLocalAddressing=ipv6\n\n[DHCP]\nRouteMetric=600\nUseMTU=true\n'
ND_DHCP = '[Match]\nName=%s\n\n[Network]\nDHCP=%s\nLinkLocalAddressing=ipv6%s\n\n[DHCP]\nRouteMetric=100\nUseMTU=%s\n'
ND_DHCP4 = ND_DHCP % ('%s', 'ipv4', '', 'true')
ND_DHCP4_NOMTU = ND_DHCP % ('%s', 'ipv4', '', 'false')
ND_DHCP6 = ND_DHCP % ('%s', 'ipv6', '', 'true')
ND_DHCP6_NOMTU = ND_DHCP % ('%s', 'ipv6', '', 'false')
ND_DHCP6_WOCARRIER = ND_DHCP % ('%s', 'ipv6', '\nConfigureWithoutCarrier=yes', 'true')
ND_DHCPYES = ND_DHCP % ('%s', 'yes', '', 'true')
ND_DHCPYES_NOMTU = ND_DHCP % ('%s', 'yes', '', 'false')
_OVS_BASE = '[Unit]\nDescription=OpenVSwitch configuration for %(iface)s\nDefaultDependencies=no\n\
Wants=ovsdb-server.service\nAfter=ovsdb-server.service\n'
OVS_PHYSICAL = _OVS_BASE + 'Requires=sys-subsystem-net-devices-%(iface)s.device\nAfter=sys-subsystem-net-devices-%(iface)s\
.device\nAfter=netplan-ovs-cleanup.service\nBefore=network.target\nWants=network.target\n%(extra)s'
OVS_VIRTUAL = _OVS_BASE + 'After=netplan-ovs-cleanup.service\nBefore=network.target\nWants=network.target\n%(extra)s'
OVS_BR_DEFAULT = 'ExecStart=' + OVS_VSCTL_PATH + ' set Bridge %(iface)s external-ids:netplan=\"true\"\n\
ExecStart=' + OVS_VSCTL_PATH + ' set-fail-mode %(iface)s standalone\nExecStart=' + OVS_VSCTL_PATH + ' set Bridge %(iface)s \
external-ids:netplan/global/set-fail-mode=\"standalone\"\nExecStart=' + OVS_VSCTL_PATH + ' set Bridge %(iface)s \
mcast_snooping_enable=false\nExecStart=' + OVS_VSCTL_PATH + ' set Bridge %(iface)s external-ids:netplan/mcast_snooping_enable=\
"false"\nExecStart=' + OVS_VSCTL_PATH + ' set Bridge %(iface)s rstp_enable=false\nExecStart=' + OVS_VSCTL_PATH + ' \
set Bridge %(iface)s external-ids:netplan/rstp_enable=\"false\"\n'
OVS_BR_EMPTY = _OVS_BASE + 'After=netplan-ovs-cleanup.service\nBefore=network.target\nWants=network.target\n\n[Service]\n\
Type=oneshot\nTimeoutStartSec=10s\nExecStart=' + OVS_VSCTL_PATH + ' --may-exist add-br %(iface)s\n' + OVS_BR_DEFAULT
OVS_CLEANUP = _OVS_BASE + 'ConditionFileIsExecutable=' + OVS_VSCTL_PATH + '\nBefore=network.target\nWants=network.target\n\n\
[Service]\nType=oneshot\nTimeoutStartSec=10s\nStartLimitBurst=0\nExecStart=/usr/sbin/netplan apply --only-ovs-cleanup\n'
UDEV_MAC_RULE = 'SUBSYSTEM=="net", ACTION=="add", DRIVERS=="%s", ATTR{address}=="%s", NAME="%s"\n'
UDEV_NO_MAC_RULE = 'SUBSYSTEM=="net", ACTION=="add", DRIVERS=="%s", NAME="%s"\n'
UDEV_SRIOV_RULE = 'ACTION=="add", SUBSYSTEM=="net", ATTRS{sriov_totalvfs}=="?*", RUN+="/usr/sbin/netplan apply --sriov-only"\n'
ND_WITHIPGW = '[Match]\nName=%s\n\n[Network]\nLinkLocalAddressing=ipv6\nAddress=%s\nAddress=%s\nGateway=%s\n\
ConfigureWithoutCarrier=yes\n'
NM_WG = '[connection]\nid=netplan-wg0\ntype=wireguard\ninterface-name=wg0\n\n[wireguard]\nprivate-key=%s\nlisten-port=%s\n%s\
\n\n[ipv4]\nmethod=manual\naddress1=15.15.15.15/24\ngateway=20.20.20.21\n\n[ipv6]\nmethod=manual\naddress1=\
2001:de:ad:be:ef:ca:fe:1/128\nip6-privacy=0\n'
ND_WG = '[NetDev]\nName=wg0\nKind=wireguard\n\n[WireGuard]\nPrivateKey%s\nListenPort=%s\n%s\n'
ND_VLAN = '[NetDev]\nName=%s\nKind=vlan\n\n[VLAN]\nId=%d\n'
ND_VXLAN = '[NetDev]\nName=%s\nKind=vxlan\n\n[VXLAN]\nVNI=%d\n'
ND_VRF = '[NetDev]\nName=%s\nKind=vrf\n\n[VRF]\nTable=%d\n'
ND_DUMMY = '[NetDev]\nName=%s\nKind=dummy\n'        # wokeignore:rule=dummy
ND_VETH = '[NetDev]\nName=%s\nKind=veth\n\n[Peer]\nName=%s\n'
SD_WPA = '''[Unit]
Description=WPA supplicant for netplan %(iface)s
DefaultDependencies=no
Requires=sys-subsystem-net-devices-%(iface)s.device
After=sys-subsystem-net-devices-%(iface)s.device
Before=network.target
Wants=network.target

[Service]
Type=simple
ExecStart=/sbin/wpa_supplicant -c /run/netplan/wpa-%(iface)s.conf -i%(iface)s -D%(drivers)s
ExecReload=/sbin/wpa_cli -i %(iface)s reconfigure
'''
NM_MANAGED = 'SUBSYSTEM=="net", ACTION=="add|change|move", ENV{ID_NET_NAME}=="%s", ENV{NM_UNMANAGED}="0"\n'
NM_UNMANAGED = 'SUBSYSTEM=="net", ACTION=="add|change|move", ENV{ID_NET_NAME}=="%s", ENV{NM_UNMANAGED}="1"\n'
NM_MANAGED_MAC = 'SUBSYSTEM=="net", ACTION=="add|change|move", ATTR{address}=="%s", ENV{NM_UNMANAGED}="0"\n'
NM_UNMANAGED_MAC = 'SUBSYSTEM=="net", ACTION=="add|change|move", ATTR{address}=="%s", ENV{NM_UNMANAGED}="1"\n'
NM_MANAGED_DRIVER = 'SUBSYSTEM=="net", ACTION=="add|change|move", ENV{ID_NET_DRIVER}=="%s", ENV{NM_UNMANAGED}="0"\n'
NM_UNMANAGED_DRIVER = 'SUBSYSTEM=="net", ACTION=="add|change|move", ENV{ID_NET_DRIVER}=="%s", ENV{NM_UNMANAGED}="1"\n'

WOKE_REPLACE_REGEX = ' +# wokeignore:rule=[a-z]+'


class NetplanV2Normalizer():

    def __init__(self):
        self.YAML_FALSE = ['n', 'no', 'off', 'false']
        self.YAML_TRUE = ['y', 'yes', 'on', 'true']
        self.DEFAULT_STANZAS = [
            'dhcp4-overrides: {}',  # 2nd level default (containing defaults itself)
            'dhcp6-overrides: {}',  # 2nd level default (containing defaults itself)
            'hidden: false',  # access-point
            'on-link: false',  # route
            'stp: true',  # paramters
            'type: unicast',  # route
            'version: 2',  # global
        ]
        self.DEFAULT_NETDEF = {
            'dhcp4': self.YAML_FALSE,
            'dhcp6': self.YAML_FALSE,
            'dhcp-identifier': ['duid'],
            'hidden': self.YAML_FALSE,
        }
        self.DEFAULT_DHCP = {
            'send-hostname': self.YAML_TRUE,
            'use-dns': self.YAML_TRUE,
            'use-hostname': self.YAML_TRUE,
            'use-mtu': self.YAML_TRUE,
            'use-ntp': self.YAML_TRUE,
            'use-routes': self.YAML_TRUE,
        }

    def _clear_mapping_defaults(self, keys, defaults, data):
        potential_defaults = list(set(keys) & set(defaults.keys()))
        for k in potential_defaults:
            if any(map(str(data[k]).lower().__eq__, defaults[k])):
                del data[k]

    def normalize_yaml_line(self, line):
        '''Process formatted YAML line by line (one setting/key per line)

        Deleting default values and re-writing to default wording
        '''
        kv = line.replace('"', '').replace('\'', '').split(':', 1)
        if len(kv) != 2 or kv[1].isspace() or kv[1] == '':
            return line  # no normalization needed; no value given

        # normalize key
        key = kv[0]
        if 'gratuitious-arp' in key:  # historically supported typo
            kv[0] = key.replace('gratuitious-arp', 'gratuitous-arp')

        # normalize value
        val = kv[1].strip()
        if val in self.YAML_FALSE:
            kv[1] = 'false'
        elif val in self.YAML_TRUE:
            kv[1] = 'true'
        elif val == '5G':
            kv[1] = '5GHz'
        elif val == '2.4G':
            kv[1] = '2.4GHz'
        else:  # no normalization needed or known
            kv[1] = val

        return ': '.join(kv)

    def normalize_yaml_tree(self, data, full_key=''):
        '''Walk the YAML dict/tree @data and sort its sequences in place

        Keeping track of the @full_key (path), e.g.: "network:ethernets:eth0:dhcp4"
        And normalizing certain netplan special cases
        '''
        if isinstance(data, list):
            scalars_only = not any(list(map(lambda elem: (isinstance(elem, dict) or isinstance(elem, list)), data)))
            # sort sequence alphabetically
            if scalars_only:
                data.sort()
                # remove duplicates (if needed)
                unique = set(data)
                if len(data) > len(unique):
                    rm_idx = set()
                    last_idx = 0
                    for elem in unique:
                        if data.count(elem) > 1:
                            idx = data.index(elem, last_idx)
                            rm_idx.add(idx)
                            last_idx = idx
                    for idx in rm_idx:
                        del data[idx]
        elif isinstance(data, dict):
            keys = data.keys()
            # expand special short forms
            if 'password' in keys and ':auth' not in full_key:
                data['auth'] = {'key-management': 'psk', 'password': data['password']}
                del data['password']
            if 'auth' in keys and data['auth'] == {}:
                data['auth'] = {'key-management': 'none'}
            # remove default stanza ("link-local: [ ipv6 ]"")
            if 'link-local' in keys and data['link-local'] == ['ipv6']:
                del data['link-local']
            # remove default stanza ("wakeonwlan: [ default ]")
            if 'wakeonwlan' in keys and data['wakeonwlan'] == ['default']:
                del data['wakeonwlan']
            # remove explicit openvswitch stanzas, they might not always be
            # defined in the original YAML (due to being implicit)
            if ('openvswitch' in keys and data['openvswitch'] == {} and
                    any(map(full_key.__contains__, [':bonds:', ':bridges:', ':vlans:']))):
                del data['openvswitch']
            # remove default empty bond-parameters, those are not rendered by the YAML generator
            if 'parameters' in keys and data['parameters'] == {} and ':bonds:' in full_key:
                del data['parameters']
            # remove default mode=infrastructore from wifi APs, keeping the SSID
            if 'mode' in keys and ':wifis:' in full_key and 'infrastructure' in data['mode']:
                del data['mode']
            # ignore renderer: on other than global levels for now, as that
            # information is currently not stored in the netdef data structure
            if ('renderer' in keys and len(full_key.split(':')) > 1 and
                    data['renderer'] in ['networkd', 'NetworkManager']):
                del data['renderer']
            # remove default values from the dhcp4/6-overrides mappings
            if full_key.endswith(':dhcp4-overrides') or full_key.endswith(':dhcp6-overrides'):
                self._clear_mapping_defaults(keys, self.DEFAULT_DHCP, data)
            # remove default values from netdef/interface mappings
            if len(full_key.split(':')) == 3:  # netdef level
                self._clear_mapping_defaults(keys, self.DEFAULT_NETDEF, data)

            # continue to walk the dict
            for key in data.keys():
                full_key_next = ':'.join([str(full_key), str(key)]) if full_key != '' else key
                self.normalize_yaml_tree(data[key], full_key_next)

    def normalize_yaml(self, yaml_dict):
        # 1st pass: normalize the YAML tree in place, sorting and removing some values
        self.normalize_yaml_tree(yaml_dict)
        # 2nd pass: sort the mapping keys and output a formatted yaml (one key per line)
        formatted_yaml = yaml.dump(yaml_dict, sort_keys=True)
        # 3rd pass: normalize the wording of certain keys/values per line
        #           and remove any line, containg only default values
        output = []
        for line in formatted_yaml.splitlines():
            line = self.normalize_yaml_line(line)
            if line.strip() in self.DEFAULT_STANZAS:
                continue
            output.append(line)
        return output


class TestBase(unittest.TestCase):

    def setUp(self):
        self.workdir = tempfile.TemporaryDirectory()
        self.confdir = os.path.join(self.workdir.name, 'etc', 'netplan')
        self.rundir = os.path.join(self.workdir.name, 'run', 'netplan')
        self.nm_enable_all_conf = os.path.join(
            self.workdir.name, 'run', 'NetworkManager', 'conf.d', '10-globally-managed-devices.conf')
        self.maxDiff = None

        # Set up a systemd generator for Netplan
        self.sd_generator = os.path.join(self.workdir.name, 'usr', 'lib', 'systemd', 'system-generators', 'netplan')
        os.makedirs(os.path.dirname(self.sd_generator))
        os.symlink(exe_generate, self.sd_generator)
        self.generator_dir = os.path.join(self.workdir.name, 'run', 'systemd', 'generator')
        self.generator_early_dir = os.path.join(self.workdir.name, 'run', 'systemd', 'generator.early')
        self.generator_late_dir = os.path.join(self.workdir.name, 'run', 'systemd', 'generator.late')

    def validate_generated_yaml(self, yaml_input):
        '''Validate a list of YAML input files one by one.

        Go through the list @yaml_input one by one, parse the YAML and
        re-generate the YAML output. Afterwards, normalize and compare the
        original (and normalized) input with the generated (and normalized)
        output.
        '''

        for input in yaml_input:
            parser = netplan.Parser()
            parser.load_yaml(input)
            state = netplan.State()
            state.import_parser_results(parser)

            # TODO: Allow handling of full hierarchy overrides,
            #       dealing only with the current element of 'yaml_input'.
            #       E.g. allow vlan.id & vlan.link to be defined in a base file.
            #       See test_routing.py:
            #       test_add_routes_to_different_tables_from_multiple_files
            #       test_add_duplicate_routes_from_multiple_files

            # Read output of the YAML generator (if any)
            output_fd = StringIO()
            state._dump_yaml(output_fd)
            output_yaml = yaml.safe_load(output_fd.getvalue())

            # Read input YAML file, as defined by the self.generate('...') method
            input_yaml = None
            with open(input, 'r') as orig:
                input_yaml = yaml.safe_load(orig.read())
                # Consider 'network: {}' and 'network: {version: 2}' to be empty
                if input_yaml is None or input_yaml == {'network': {}} or input_yaml == {'network': {'version': 2}}:
                    input_yaml = yaml.safe_load('')

            # Normalize input and output YAML
            netplan_normalizer = NetplanV2Normalizer()
            input_lines = netplan_normalizer.normalize_yaml(input_yaml)
            output_lines = netplan_normalizer.normalize_yaml(output_yaml)

            # Check if (normalized) input and (normalized) output are equal
            yaml_files_differ = len(input_lines) != len(output_lines)
            if not yaml_files_differ:  # pragma: no cover (only execited in error case)
                for i in range(len(input_lines)):
                    if input_lines[i] != output_lines[i]:
                        yaml_files_differ = True
                        break
            if yaml_files_differ:  # pragma: no cover (only execited in error case)
                fromfile = 'original (%s)' % input
                for line in difflib.unified_diff(input_lines, output_lines, fromfile, tofile='generated', lineterm=''):
                    print(line, flush=True)
                self.fail('Re-generated YAML file does not match (adopt netplan.c YAML generator?)')

    def generate(self, yaml, expect_fail=False, extra_args=[], confs=None, skip_generated_yaml_validation=False,
                 ignore_errors=False):
        '''Call generate with given YAML string as configuration

        Return stderr output.
        '''
        yaml_input = []
        conf = os.path.join(self.confdir, 'a.yaml')
        os.makedirs(os.path.dirname(conf), exist_ok=True)
        if yaml is not None:
            with open(conf, 'w') as f:
                f.write(yaml)
            os.chmod(conf, mode=0o600)
            yaml_input.append(conf)
        if confs:
            for f, contents in confs.items():
                path = os.path.join(self.confdir, f + '.yaml')
                with open(path, 'w') as f:
                    f.write(contents)
                os.chmod(path, mode=0o600)
                yaml_input.append(path)

        # Generators executed by the system manager are invoked in a sandbox
        # where most of the file system is read-only except for the generator
        # output directories. We considered restricting PrivateTmp=disconnected,
        # too, but it conflicts with self.workdir in public /tmp.
        # https://www.freedesktop.org/software/systemd/man/latest/systemd.generator.html#Description
        os.makedirs(self.generator_dir, mode=0o755, exist_ok=True)
        os.makedirs(self.generator_early_dir, mode=0o755, exist_ok=True)
        os.makedirs(self.generator_late_dir, mode=0o755, exist_ok=True)
        sandbox = ['systemd-run', '--user', '--pipe', '--collect', '--quiet',
                   '--setenv=LD_LIBRARY_PATH=.:{}'.format(os.environ.get('LD_LIBRARY_PATH')),
                   '--setenv=G_DEBUG=fatal-criticals',
                   '--property=ProtectSystem=strict',
                   '--property=ReadWritePaths=' + self.generator_dir,
                   '--property=ReadWritePaths=' + self.generator_early_dir,
                   '--property=ReadWritePaths=' + self.generator_late_dir,
                   # Allow writing of test coverage data inside the project root (meson's _build-cov/ dir)
                   '--property=ReadWritePaths=' + os.path.dirname(os.environ.get('COVERAGE_PROCESS_START', '/home/'))]

        if os.environ.get('NETPLAN_PARSER_IGNORE_ERRORS'):
            del os.environ['NETPLAN_PARSER_IGNORE_ERRORS']
        if not ignore_errors:
            # for testing only
            os.environ.setdefault('NETPLAN_PARSER_IGNORE_ERRORS', '0')
            sandbox += ['--setenv=NETPLAN_PARSER_IGNORE_ERRORS={}'
                        .format(os.environ.get('NETPLAN_PARSER_IGNORE_ERRORS'))]

        # Avoid the systemd-run sandbox when run in chroot (e.g. during sbuild),
        # as it does not provide a functional systemd environment. The
        # ReadWritePaths restriction will still be tested inside the upstream CI
        # and locally, using 'make check'.
        if not os.path.isdir('/run/systemd/system'):
            sandbox = []  # pragma: nocover (runs during sbuild)

        argv_gen = sandbox + [self.sd_generator, '--root-dir', self.workdir.name,
                              self.generator_dir, self.generator_early_dir, self.generator_late_dir]
        argv = [exe_configure, '--root-dir', self.workdir.name] + extra_args
        if 'TEST_SHELL' in os.environ:  # pragma nocover
            print('Test is about to run:\n%s' % ' '.join(argv))
            subprocess.call(['bash', '-i'], cwd=self.workdir.name)

        # if called as generator (passing generator{early,late} dirs) it will
        # ignore errors by default. Errors are forced via NETPLAN_PARSER_IGNORE_ERRORS=0.
        if ignore_errors:
            argv_gen += ['--ignore-errors']
            argv += ['--ignore-errors']

        g = subprocess.Popen(argv_gen, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, text=True)
        (out_gen, err_gen) = g.communicate()
        p = subprocess.Popen(argv, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, text=True)
        (out, err) = p.communicate()
        if expect_fail:
            self.assertGreater(p.returncode, 0, err)
            if not extra_args:
                self.assertGreater(g.returncode, 0, err_gen)
            else:
                # sd-generator cannot handle the extra_args and will pass
                # (for testing only)
                self.assertEqual(g.returncode, 0)
        self.assertEqual(out_gen, '')
        self.assertEqual(out, '')
        if not expect_fail and not skip_generated_yaml_validation:
            yaml_input = list(set(yaml_input + extra_args))
            yaml_input.sort()
            self.validate_generated_yaml(yaml_input)

        ret_output = err_gen if err_gen else err
        if err_gen and err:
            ret_output = err_gen + '#*#' + err

        return ret_output

    def eth_name(self):
        """Return a link name.

        Use when you need a link name for a test but don't want to
        encode a made up name in the test.
        """
        return 'eth' + ''.join(random.sample(string.ascii_letters + string.digits, k=4))

    def file_exists(self, filename, backend='systemd') -> bool:
        if backend == 'systemd':
            path = os.path.join(self.workdir.name, 'run', 'systemd', 'network', filename)
        else:
            path = os.path.join(self.workdir.name, 'run', 'NetworkManager', 'system-connections', filename)
        return os.path.exists(path)

    def assert_networkd(self, file_contents_map):
        networkd_dir = os.path.join(self.workdir.name, 'run', 'systemd', 'network')
        if not file_contents_map:
            self.assertFalse(os.path.exists(networkd_dir))
            return

        self.assertEqual(set(os.listdir(self.workdir.name)) - {'usr', 'lib'}, {'etc', 'run'})
        self.assertEqual(set(os.listdir(networkd_dir)),
                         {'10-netplan-' + f for f in file_contents_map})
        for fname, contents in file_contents_map.items():
            contents = re.sub(WOKE_REPLACE_REGEX, '', contents)
            with open(os.path.join(networkd_dir, '10-netplan-' + fname)) as f:
                self.assertEqual(f.read(), contents)

    def assert_networkd_udev(self, file_contents_map):
        udev_dir = os.path.join(self.workdir.name, 'run', 'udev', 'rules.d')
        if not file_contents_map:
            # it can either not exist, or can only contain 90-netplan.rules
            self.assertTrue((not os.path.exists(udev_dir)) or
                            (os.listdir(udev_dir) == ['90-netplan.rules']))
            return

        self.assertEqual(set(os.listdir(udev_dir)) - set(['90-netplan.rules']),
                         {'99-netplan-' + f for f in file_contents_map})
        for fname, contents in file_contents_map.items():
            with open(os.path.join(udev_dir, '99-netplan-' + fname)) as f:
                self.assertEqual(f.read(), contents)

    def get_network_config_for_link(self, link_name):
        """Return the content of the .network file for `link_name`."""
        networkd_dir = os.path.join(self.workdir.name, 'run', 'systemd', 'network')
        with open(os.path.join(networkd_dir, '10-netplan-{}.network'.format(link_name))) as f:
            return f.read()

    def get_optional_addresses(self, eth_name):
        config = self.get_network_config_for_link(eth_name)
        r = set()
        prefix = "OptionalAddresses="
        for line in config.splitlines():
            if line.startswith(prefix):
                r.add(line[len(prefix):])
        return r

    def assert_wpa_supplicant(self, iface, content):
        conf_path = os.path.join(self.workdir.name, 'run', 'netplan', "wpa-" + iface + ".conf")
        with open(conf_path) as f:
            self.assertEqual(f.read(), content)

    def assert_nm(self, connections_map=None, conf=None):
        # check config
        conf_path = os.path.join(self.workdir.name, 'run', 'NetworkManager', 'conf.d', 'netplan.conf')
        if conf:
            conf = re.sub(WOKE_REPLACE_REGEX, '', conf)
            with open(conf_path) as f:
                self.assertEqual(f.read(), conf)
        else:
            if os.path.exists(conf_path):
                with open(conf_path) as f:  # pragma: nocover
                    self.fail('unexpected %s:\n%s' % (conf_path, f.read()))

        # check connections
        con_dir = os.path.join(self.workdir.name, 'run', 'NetworkManager', 'system-connections')
        if connections_map:
            self.assertEqual(set(os.listdir(con_dir)),
                             set(['netplan-' + n.split('.nmconnection')[0] + '.nmconnection' for n in connections_map]))
            for fname, contents in connections_map.items():
                contents = re.sub(WOKE_REPLACE_REGEX, '', contents)
                extension = ''
                if '.nmconnection' not in fname:
                    extension = '.nmconnection'
                with open(os.path.join(con_dir, 'netplan-' + fname + extension)) as f:
                    self.assertEqual(f.read(), contents)
                    # NM connection files might contain secrets
                    self.assertEqual(stat.S_IMODE(os.fstat(f.fileno()).st_mode), 0o600)
        else:
            if os.path.exists(con_dir):
                self.assertEqual(os.listdir(con_dir), [])

    def assert_nm_udev(self, contents):
        rule_path = os.path.join(self.workdir.name, 'run/udev/rules.d/90-netplan.rules')
        if contents is None:
            self.assertFalse(os.path.exists(rule_path))
            return
        with open(rule_path) as f:
            lines = []
            for line in f.readlines():
                # ignore any comment in udev rules.d file
                if not line.startswith('#'):
                    lines.append(line)
            self.assertEqual(''.join(lines), contents)

    def assert_ovs(self, file_contents_map):
        systemd_dir = self.generator_late_dir
        if not file_contents_map:
            # in this case we assume no OVS configuration should be present
            self.assertFalse(glob.glob(os.path.join(systemd_dir, '*netplan-ovs-*.service')))
            return

        self.assertIn('run', os.listdir(self.workdir.name))
        ovs_systemd_dir = set(os.listdir(systemd_dir))
        ovs_systemd_dir.remove('systemd-networkd.service.wants')
        if 'systemd-networkd-wait-online.service.d' in ovs_systemd_dir:
            ovs_systemd_dir.remove('systemd-networkd-wait-online.service.d')
        self.assertEqual(ovs_systemd_dir, {'netplan-ovs-' + f for f in file_contents_map})
        for fname, contents in file_contents_map.items():
            fname = 'netplan-ovs-' + fname
            with open(os.path.join(systemd_dir, fname)) as f:
                self.assertEqual(f.read(), contents)
            if fname.endswith('.service'):
                link_path = os.path.join(
                    systemd_dir, 'systemd-networkd.service.wants', fname)
                self.assertTrue(os.path.islink(link_path))
                link_target = os.readlink(link_path)
                self.assertEqual(link_target,
                                 os.path.join(
                                    self.generator_late_dir, fname))

    def assert_sriov(self, file_contents_map):
        systemd_dir = self.generator_late_dir
        sriov_systemd_dir = glob.glob(os.path.join(systemd_dir, '*netplan-sriov-*.service'))
        self.assertEqual(set(os.path.basename(file) for file in sriov_systemd_dir),
                         {'netplan-sriov-' + f for f in file_contents_map})
        self.assertEqual(set(os.listdir(self.workdir.name)) - {'usr', 'lib'}, {'etc', 'run'})

        for file in sriov_systemd_dir:
            basename = os.path.basename(file)
            with open(file, 'r') as f:
                contents = f.read()
                map_contents = file_contents_map.get(basename.replace('netplan-sriov-', ''))
                self.assertEqual(map_contents, contents)



#
# Tests for ethernet devices config generated via netplan
#
# Copyright (C) 2018 Canonical, Ltd.
# Author: Mathieu Trudel-Lapierre <mathieu.trudel.lapierre@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from .base import TestBase, ND_DHCP4, UDEV_MAC_RULE, UDEV_NO_MAC_RULE, \
    NM_MANAGED, NM_UNMANAGED, NM_MANAGED_MAC, NM_UNMANAGED_MAC, \
    NM_MANAGED_DRIVER, NM_UNMANAGED_DRIVER


class TestNetworkd(TestBase):

    def test_eth_wol(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth0:
      wakeonlan: true
      dhcp4: n''')

        self.assert_networkd({'eth0.link': '[Match]\nOriginalName=eth0\n\n[Link]\nWakeOnLan=magic\n',
                              'eth0.network': '''[Match]
Name=eth0

[Network]
LinkLocalAddressing=ipv6
'''})
        self.assert_networkd_udev(None)
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % 'eth0')
        # should not allow NM to manage everything
        self.assertFalse(os.path.exists(self.nm_enable_all_conf))

    def test_eth_lldp(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth0:
      dhcp4: n
      emit-lldp: true''')

        self.assert_networkd({'eth0.network': '''[Match]
Name=eth0

[Network]
EmitLLDP=true
LinkLocalAddressing=ipv6
'''})

    def test_eth_mtu(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth1:
      mtu: 1280
      dhcp4: n''')

        self.assert_networkd({'eth1.link': '[Match]\nOriginalName=eth1\n\n[Link]\nWakeOnLan=off\nMTUBytes=1280\n',
                              'eth1.network': '''[Match]
Name=eth1

[Link]
MTUBytes=1280

[Network]
LinkLocalAddressing=ipv6
'''})
        self.assert_networkd_udev(None)

    def test_eth_match_by_driver_rename(self):
        self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        driver: ixgbe
      set-name: lom1''')

        self.assert_networkd({'def1.link': '[Match]\nDriver=ixgbe\n\n[Link]\nName=lom1\nWakeOnLan=off\n',
                              'def1.network': '''[Match]
Driver=ixgbe
Name=lom1

[Network]
LinkLocalAddressing=ipv6
'''})
        self.assert_networkd_udev({'def1.rules': (UDEV_NO_MAC_RULE % ('ixgbe', 'lom1'))})
        # NM cannot match by driver, so denylisting needs to happen via udev
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % 'lom1' + NM_UNMANAGED_DRIVER % 'ixgbe')

    def test_eth_match_by_mac_rename(self):
        self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        macaddress: 11:22:33:44:55:66
      set-name: lom1''')

        self.assert_networkd({'def1.link': '[Match]\nPermanentMACAddress=11:22:33:44:55:66\n\n[Link]\nName=lom1\nWakeOnLan=off\n',
                              'def1.network': '''[Match]
PermanentMACAddress=11:22:33:44:55:66
Name=lom1

[Network]
LinkLocalAddressing=ipv6
'''})
        self.assert_networkd_udev({'def1.rules': (UDEV_MAC_RULE % ('?*', '11:22:33:44:55:66', 'lom1'))})
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % 'lom1' + NM_UNMANAGED_MAC % '11:22:33:44:55:66')

    # https://bugs.launchpad.net/netplan/+bug/1848474
    def test_eth_match_by_mac_infiniband(self):
        self.generate('''network:
  version: 2
  ethernets:
    ib0:
      match:
        macaddress: 11:22:33:44:55:66:77:88:99:00:11:22:33:44:55:66:77:88:99:00
      dhcp4: true
      infiniband-mode: connected''')

        self.assert_networkd({'ib0.network': '''[Match]
PermanentMACAddress=11:22:33:44:55:66:77:88:99:00:11:22:33:44:55:66:77:88:99:00

[Network]
DHCP=ipv4
LinkLocalAddressing=ipv6

[DHCP]
RouteMetric=100
UseMTU=true

[IPoIB]
Mode=connected
'''})
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED_MAC % '11:22:33:44:55:66:77:88:99:00:11:22:33:44:55:66:77:88:99:00')

    def test_eth_implicit_name_match_dhcp4(self):
        self.generate('''network:
  version: 2
  ethernets:
    engreen:
      dhcp4: y''')

        self.assert_networkd({'engreen.network': ND_DHCP4 % 'engreen'})
        self.assert_networkd_udev(None)

    def test_eth_match_dhcp4(self):
        self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        driver: ixgbe
      dhcp4: true''')

        self.assert_networkd({'def1.network': '''[Match]
Driver=ixgbe

[Network]
DHCP=ipv4
LinkLocalAddressing=ipv6

[DHCP]
RouteMetric=100
UseMTU=true
'''})
        self.assert_networkd_udev(None)
        self.assert_nm_udev(NM_UNMANAGED_DRIVER % 'ixgbe')

    def test_eth_match_name(self):
        self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        name: green
      dhcp4: true''')

        self.assert_networkd({'def1.network': ND_DHCP4 % 'green'})
        self.assert_networkd_udev(None)
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % 'green')

    def test_eth_set_mac(self):
        self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        name: green
      macaddress: 00:01:02:03:04:05
      dhcp4: true''')

        self.assert_networkd({'def1.network': (ND_DHCP4 % 'green')
                              .replace('[Network]', '[Link]\nMACAddress=00:01:02:03:04:05\n\n[Network]')
                              })
        self.assert_networkd_udev(None)

    def test_eth_set_mac_value_not_valid(self):
        res = self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        name: green
      macaddress: typorandom
      dhcp4: true''', expect_fail=True)

        self.assertIn('Invalid MAC address \'typorandom\'', res)

    def test_eth_set_mac_value_only_valid_for_network_manager(self):
        res = self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        name: green
      macaddress: stable
      dhcp4: true''', expect_fail=True)

        self.assertIn('networkd backend does not support the MAC address option \'stable\'', res)

    def test_eth_set_mac_special_values(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth0:
      macaddress: permanent
      dhcp4: true
    eth1:
      macaddress: random
      dhcp4: true''')

        self.assert_networkd({'eth0.link': '[Match]\nOriginalName=eth0\n\n[Link]\nWakeOnLan=off\nMACAddressPolicy=persistent\n',
                              'eth0.network': ND_DHCP4 % 'eth0',
                              'eth1.link': '[Match]\nOriginalName=eth1\n\n[Link]\nWakeOnLan=off\nMACAddressPolicy=random\n',
                              'eth1.network': ND_DHCP4 % 'eth1'})

    def test_eth_match_name_rename(self):
        self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        name: green
      set-name: blue
      dhcp4: true''')

        # the .network needs to match on the renamed name
        self.assert_networkd({'def1.link': '[Match]\nOriginalName=green\n\n[Link]\nName=blue\nWakeOnLan=off\n',
                              'def1.network': ND_DHCP4 % 'blue'})

        # The udev rules engine does support renaming by name
        self.assert_networkd_udev(None)

        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % 'blue' + NM_UNMANAGED % 'green')

    def test_eth_match_all_names(self):
        self.generate('''network:
  version: 2
  ethernets:
    def1:
      match: {name: "*"}
      dhcp4: true''')

        self.assert_networkd({'def1.network': ND_DHCP4 % '*'})
        self.assert_networkd_udev(None)
        self.assert_nm(None)
        self.assert_nm_udev(NM_UNMANAGED % '*')

    def test_eth_match_all(self):
        self.generate('''network:
  version: 2
  ethernets:
    def1:
      match: {}
      dhcp4: true''')

        self.assert_networkd({'def1.network': '[Match]\n\n[Network]\nDHCP=ipv4\nLinkLocalAddressing=ipv6\n\n'
                                              '[DHCP]\nRouteMetric=100\nUseMTU=true\n'})
        self.assert_networkd_udev(None)
        self.assert_nm(None, '''[device-netplan.ethernets.def1]
match-device=type:ethernet
managed=0\n\n''')
        self.assert_nm_udev(None)

    def test_match_multiple(self):
        self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        name: en1s*
        macaddress: 00:11:22:33:44:55
      dhcp4: on''')
        self.assert_networkd({'def1.network': '''[Match]
PermanentMACAddress=00:11:22:33:44:55
Name=en1s*

[Network]
DHCP=ipv4
LinkLocalAddressing=ipv6

[DHCP]
RouteMetric=100
UseMTU=true
'''})
        self.assert_nm(None)
        self.assert_nm_udev('SUBSYSTEM=="net", ACTION=="add|change|move", ENV{ID_NET_NAME}=="en1s*", '
                            'ATTR{address}=="00:11:22:33:44:55", ENV{NM_UNMANAGED}="1"\n')


class TestNetworkManager(TestBase):

    def test_eth_wol(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    eth0:
      wakeonlan: true''')

        self.assert_nm({'eth0': '''[connection]
id=netplan-eth0
type=ethernet
interface-name=eth0

[ethernet]
wake-on-lan=1

[ipv4]
method=link-local

[ipv6]
method=ignore
'''})
        # should allow NM to manage everything else
        self.assertTrue(os.path.exists(self.nm_enable_all_conf))
        self.assert_networkd({'eth0.link': '[Match]\nOriginalName=eth0\n\n[Link]\nWakeOnLan=magic\n'})
        self.assert_nm_udev(NM_MANAGED % 'eth0')

    def test_eth_mtu(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    eth1:
      mtu: 1280
      dhcp4: n''')

        self.assert_networkd({'eth1.link': '[Match]\nOriginalName=eth1\n\n[Link]\nWakeOnLan=off\nMTUBytes=1280\n'})
        self.assert_nm({'eth1': '''[connection]
id=netplan-eth1
type=ethernet
interface-name=eth1

[ethernet]
wake-on-lan=0
mtu=1280

[ipv4]
method=link-local

[ipv6]
method=ignore
'''})

    def test_eth_set_mac(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    eth0:
      macaddress: 00:01:02:03:04:05
      dhcp4: true''')

        self.assert_networkd(None)

        self.assert_nm({'eth0': '''[connection]
id=netplan-eth0
type=ethernet
interface-name=eth0

[ethernet]
wake-on-lan=0
cloned-mac-address=00:01:02:03:04:05

[ipv4]
method=auto

[ipv6]
method=ignore
'''})

    def test_eth_set_mac_special_values(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    eth0:
      macaddress: preserve
      dhcp4: true
    eth1:
      macaddress: permanent
      dhcp4: true
    eth2:
      macaddress: random
      dhcp4: true
    eth3:
      macaddress: stable
      dhcp4: true''')

        self.assert_networkd(None)

        self.assert_nm({'eth0': '''[connection]
id=netplan-eth0
type=ethernet
interface-name=eth0

[ethernet]
wake-on-lan=0
cloned-mac-address=preserve

[ipv4]
method=auto

[ipv6]
method=ignore
''',
                       'eth1': '''[connection]
id=netplan-eth1
type=ethernet
interface-name=eth1

[ethernet]
wake-on-lan=0
cloned-mac-address=permanent

[ipv4]
method=auto

[ipv6]
method=ignore
''',
                        'eth2': '''[connection]
id=netplan-eth2
type=ethernet
interface-name=eth2

[ethernet]
wake-on-lan=0
cloned-mac-address=random

[ipv4]
method=auto

[ipv6]
method=ignore
''',
                        'eth3': '''[connection]
id=netplan-eth3
type=ethernet
interface-name=eth3

[ethernet]
wake-on-lan=0
cloned-mac-address=stable

[ipv4]
method=auto

[ipv6]
method=ignore
'''})

    def test_eth_set_mac_special_values_error(self):
        res = self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    eth0:
      macaddress: preservetypo
      dhcp4: true''', expect_fail=True)

        error = ("Invalid MAC address 'preservetypo', must be XX:XX:XX:XX:XX:XX, "
                 "XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX or "
                 "one of 'permanent', 'random', 'stable', 'preserve', 'stable-ssid' (Wi-Fi only)")
        self.assertIn(error, res)

    def test_eth_set_mac_special_values_ethernet_stable_ssid(self):
        res = self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    eth0:
      macaddress: stable-ssid
      dhcp4: true''', expect_fail=True)

        error = ("Invalid MAC address 'stable-ssid', must be XX:XX:XX:XX:XX:XX, "
                 "XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX:XX or "
                 "one of 'permanent', 'random', 'stable', 'preserve', 'stable-ssid' (Wi-Fi only)")
        self.assertIn(error, res)

    def test_eth_match_by_driver(self):
        err = self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    def1:
      match:
        driver: ixgbe''', expect_fail=True)
        self.assertIn('NetworkManager definitions do not support matching by driver', err)

    def test_eth_match_by_drivers(self):
        self.generate('''network:
  version: 2
  renderer: networkd
  ethernets:
    def1:
      match:
        driver: ["bcmgenet", "smsc*"]''')
        self.assert_networkd({'def1.network': '''[Match]
Driver=bcmgenet smsc*

[Network]
LinkLocalAddressing=ipv6
'''})

    def test_eth_match_by_drivers_whitespace(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        driver: "bcmgenet smsc*"''', expect_fail=True)
        self.assertIn('A \'driver\' glob cannot contain whitespace', err)

    def test_eth_match_by_drivers_whitespace_sequence(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        driver: ["ixgbe", "bcmgenet smsc*"]''', expect_fail=True)
        self.assertIn('A \'driver\' glob cannot contain whitespace', err)

    def test_eth_match_by_drivers_invalid_sequence(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        driver: []''', expect_fail=True)
        self.assertIn('invalid sequence for \'driver\'', err)

    def test_eth_match_by_drivers_invalid_type(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    def1:
      match:
        driver:
            some_mapping: true''', expect_fail=True)
        self.assertIn('invalid type for \'driver\': must be a scalar or a sequence of scalars', err)

    def test_eth_match_by_driver_rename(self):
        # in this case udev will rename the device so that NM can use the name
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    def1:
      match:
        driver: ixgbe
      set-name: lom1''')

        self.assert_networkd({'def1.link': '[Match]\nDriver=ixgbe\n\n[Link]\nName=lom1\nWakeOnLan=off\n'})
        self.assert_networkd_udev({'def1.rules': (UDEV_NO_MAC_RULE % ('ixgbe', 'lom1'))})
        self.assert_nm({'def1': '''[connection]
id=netplan-def1
type=ethernet
interface-name=lom1

[ethernet]
wake-on-lan=0

[ipv4]
method=link-local

[ipv6]
method=ignore
'''})
        self.assert_nm_udev(NM_MANAGED % 'lom1' + NM_MANAGED_DRIVER % 'ixgbe')

    def test_eth_match_by_mac_rename(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    def1:
      match:
        macaddress: 11:22:33:44:55:66
      set-name: lom1''')

        self.assert_networkd({'def1.link': '''[Match]
PermanentMACAddress=11:22:33:44:55:66\n\n[Link]\nName=lom1\nWakeOnLan=off\n'''})
        self.assert_networkd_udev({'def1.rules': (UDEV_MAC_RULE % ('?*', '11:22:33:44:55:66', 'lom1'))})
        self.assert_nm({'def1': '''[connection]
id=netplan-def1
type=ethernet
interface-name=lom1

[ethernet]
wake-on-lan=0

[ipv4]
method=link-local

[ipv6]
method=ignore
'''})
        self.assert_nm_udev(NM_MANAGED % 'lom1' + NM_MANAGED_MAC % '11:22:33:44:55:66')

    def test_eth_implicit_name_match_dhcp4(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    engreen:
      dhcp4: true''')

        self.assert_nm({'engreen': '''[connection]
id=netplan-engreen
type=ethernet
interface-name=engreen

[ethernet]
wake-on-lan=0

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        self.assert_networkd({})

    def test_eth_match_mac_dhcp4(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    def1:
      match:
        macaddress: 11:22:33:44:55:66
      dhcp4: true''')

        self.assert_nm({'def1': '''[connection]
id=netplan-def1
type=ethernet

[ethernet]
wake-on-lan=0
mac-address=11:22:33:44:55:66

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        self.assert_networkd({})

    def test_eth_match_name(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    def1:
      match:
        name: green
      dhcp4: true''')

        self.assert_nm({'def1': '''[connection]
id=netplan-def1
type=ethernet
interface-name=green

[ethernet]
wake-on-lan=0

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        self.assert_networkd({})
        self.assert_nm_udev(NM_MANAGED % 'green')

    def test_eth_match_name_rename(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    def1:
      match:
        name: green
      set-name: blue
      dhcp4: true''')

        # The udev rules engine does support renaming by name
        self.assert_networkd_udev(None)

        # NM needs to match on the renamed name
        self.assert_nm({'def1': '''[connection]
id=netplan-def1
type=ethernet
interface-name=blue

[ethernet]
wake-on-lan=0

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        # ... while udev renames it
        self.assert_networkd({'def1.link': '[Match]\nOriginalName=green\n\n[Link]\nName=blue\nWakeOnLan=off\n'})
        self.assert_nm_udev(NM_MANAGED % 'blue' + NM_MANAGED % 'green')

    def test_eth_match_name_glob(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    def1:
      match: {name: "en*"}
      dhcp4: true''')

        self.assert_nm({'def1': '''[connection]
id=netplan-def1
type=ethernet

[ethernet]
wake-on-lan=0

[match]
interface-name=en*;

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        self.assert_networkd({})

    def test_eth_match_all(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    def1:
      match: {}
      dhcp4: true''')

        self.assert_nm({'def1': '''[connection]
id=netplan-def1
type=ethernet

[ethernet]
wake-on-lan=0

[ipv4]
method=auto

[ipv6]
method=ignore
'''}, '''[device-netplan.ethernets.def1]
match-device=type:ethernet
managed=1\n\n''')
        self.assert_nm_udev(None)
        self.assert_networkd({})

    def test_match_multiple(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    def1:
      match:
        name: engreen
        macaddress: 00:11:22:33:44:55
      dhcp4: yes''')
        self.assert_nm({'def1': '''[connection]
id=netplan-def1
type=ethernet
interface-name=engreen

[ethernet]
wake-on-lan=0
mac-address=00:11:22:33:44:55

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        self.assert_networkd({})
        self.assert_nm_udev('SUBSYSTEM=="net", ACTION=="add|change|move", ENV{ID_NET_NAME}=="engreen", '
                            'ATTR{address}=="00:11:22:33:44:55", ENV{NM_UNMANAGED}="0"\n')

    def test_offload(self):
        self.generate('''network:
  version: 2
  ethernets:
    eth1:
      receive-checksum-offload: true
      transmit-checksum-offload: off
      tcp-segmentation-offload: true
      tcp6-segmentation-offload: false
      generic-segmentation-offload: true
      generic-receive-offload: no
      large-receive-offload: true''')

        self.assert_networkd({'eth1.link': '''[Match]
OriginalName=eth1

[Link]
WakeOnLan=off
ReceiveChecksumOffload=true
TransmitChecksumOffload=false
TCPSegmentationOffload=true
TCP6SegmentationOffload=false
GenericSegmentationOffload=true
GenericReceiveOffload=false
LargeReceiveOffload=true
''',
                              'eth1.network': '''[Match]
Name=eth1

[Network]
LinkLocalAddressing=ipv6
'''})
        self.assert_networkd_udev(None)

    def test_offload_invalid(self):
        err = self.generate('''network:
  version: 2
  ethernets:
    eth1:
      generic-receive-offload: n
      receive-checksum-offload: true
      tcp-segmentation-offload: true
      tcp6-segmentation-offload: false
      generic-segmentation-offload: true
      transmit-checksum-offload: xx
      large-receive-offload: true''', expect_fail=True)
        self.assertIn('invalid boolean value \'xx\'', err)

    # https://bugs.launchpad.net/netplan/+bug/1848474
    def test_eth_match_by_mac_infiniband(self):
        self.generate('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    ib0:
      match:
        macaddress: 11:22:33:44:55:66:77:88:99:00:11:22:33:44:55:66:77:88:99:00
      dhcp4: true
      infiniband-mode: datagram''')

        self.assert_networkd(None)
        self.assert_nm({'ib0': '''[connection]
id=netplan-ib0
type=infiniband

[infiniband]
mac-address=11:22:33:44:55:66:77:88:99:00:11:22:33:44:55:66:77:88:99:00
transport-mode=datagram

[ipv4]
method=auto

[ipv6]
method=ignore
'''})
        self.assert_nm_udev(NM_MANAGED_MAC % '11:22:33:44:55:66:77:88:99:00:11:22:33:44:55:66:77:88:99:00')
